Require Import model_facts.
Require Import models.

Set Implicit Arguments.

(* We define the notion of a data race free program with respect to a language
   model L and a memory model M. *)

Module DataRaceFree (L : LANGUAGE_MODEL) (M : MEMORY_MODEL L)
  (Sys : SYSTEM_MODEL L M).
  Module Facts := ModelFacts L M Sys.

  Definition no_imminent_races (H : M.heap) (P : L.program_state) : Prop :=
    forall H1 P1 H2 P2 H3 P3 t1 a1 t2 a2,
      Sys.system_step H1 P1 (Some (t1, a1)) H2 P2 ->
      Sys.system_step H2 P2 (Some (t2, a2)) H3 P3 ->
      (~ data_race a1 a2) \/ t1 = t2.

  Definition race_free_state (H : M.heap) (P : L.program_state) : Prop :=
    forall H' P' tr,
      Sys.system_step_many H P H' P' tr ->
      no_imminent_races H' P'.

  Definition race_free_program (P : L.program_state) : Prop :=
    race_free_state M.empty_heap P.

  Lemma step_many_race_free_state : forall H1 P1 H2 P2 tr,
    race_free_state H1 P1 ->
    Sys.system_step_many H1 P1 H2 P2 tr ->
    race_free_state H2 P2.
  Proof.
    intros H1 P1 H2 P2 tr drf steps_many.
    intros H3 P3 tr' steps_many'.
    unfold race_free_state in drf.
    eapply drf. eapply Facts.system_step_many_append; eauto.
  Qed.
End DataRaceFree.
