Require Export List.

Require Export actions.
Require Export indexes.

Set Implicit Arguments.

Module Type LANGUAGE_MODEL.
  Parameter val : Set.
  Parameter program_state : Type.
  Parameter program_step : program_state -> effect val -> program_state -> Prop.

  Inductive program_step_many : program_state -> program_state -> trace val
    -> Prop :=
  | program_step_many_refl : forall P,
    program_step_many P P nil
  | program_step_many_trans : forall P1 P2 P3 tr e,
    program_step_many P1 P2 tr ->
    program_step P2 e P3 ->
    program_step_many P1 P3 (e :: tr).

  Axiom no_empty_steps : forall P P',
    ~ (program_step P None P').
End LANGUAGE_MODEL.

Module Type MEMORY_MODEL (L : LANGUAGE_MODEL).
  Parameter heap : Type.
  Parameter empty_heap : heap.
  Parameter heap_step : heap -> effect L.val -> heap -> Prop.

  Inductive heap_step_many : heap -> heap -> trace L.val -> Prop :=
  | heap_step_many_refl : forall H,
    heap_step_many H H nil
  | heap_step_many_trans : forall H1 H2 H3 tr e,
    heap_step_many H1 H2 tr ->
    heap_step H2 e H3 ->
    heap_step_many H1 H3 (e :: tr).

  Axiom no_pure_steps : forall H H' t,
    ~ (heap_step H (Some (t, pure _)) H').
End MEMORY_MODEL.

Module Type SYSTEM_MODEL (L : LANGUAGE_MODEL) (M : MEMORY_MODEL L).
  Inductive system_step : M.heap -> L.program_state -> effect L.val
    -> M.heap -> L.program_state -> Prop :=
  | system_step_mutual : forall H1 H2 P1 P2 t a,
    L.program_step P1 (Some (t, a)) P2 ->
    M.heap_step H1 (Some (t, a)) H2 ->
    system_step H1 P1 (Some (t, a)) H2 P2
  | system_step_heap : forall H1 H2 P,
    M.heap_step H1 None H2 ->
    system_step H1 P None H2 P
  | system_step_program : forall H P1 P2 t,
    L.program_step P1 (Some (t, pure _)) P2 ->
    system_step H P1 (Some (t, pure _)) H P2.

  Inductive system_step_many : M.heap -> L.program_state -> M.heap
    -> L.program_state -> trace L.val -> Prop :=
  | system_step_many_refl : forall H P,
    system_step_many H P H P nil
  | system_step_many_trans : forall H1 P1 H2 P2 H3 P3 tr e,
    system_step_many H1 P1 H2 P2 tr ->
    system_step H2 P2 e H3 P3 ->
    system_step_many H1 P1 H3 P3 (e :: tr).
End SYSTEM_MODEL.
