Require Import Metatheory.

Require Import models.
Require Import types.

Set Implicit Arguments.

Module System_Model_Type_Preservation (Pres : TYPE_MODEL_PRESERVATION)
  (MPres : MEMORY_MODEL_TYPE_PRESERVATION Pres.Typ)
  (Sys : SYSTEM_MODEL Pres.Typ.L MPres.M).
  Import Pres.
  Import Typ.
  Import L.
  Import MPres.
  Import M.
  Import Sys.

  Module Facts := Type_Model_Facts Typ.

  Theorem preservation : forall Sigma H P e H' P',
    well_typed Sigma H ->
    typing Sigma P ->
    system_step H P e H' P' ->
    exists Sigma',
      extends Sigma Sigma' /\
      well_typed Sigma' H' /\
      typing Sigma' P'.
  Proof.
    intros Sigma H P e H' P' typing_H typing_P steps.
    inversion steps as [ ? ? ? ? t a steps_P steps_H | | ? ? ? t steps_P ];
      subst.

    (* mutual *)
    pose proof typing_step typing_H steps_H as typing_in.
    pose proof Pres.preservation typing_P steps_P typing_in
      as [ Sigma' [ typing_out typing_P' ] ].
    pose proof MPres.preservation typing_H steps_H typing_out as typing_H'.
    pose proof Facts.typing_extends typing_in typing_out as ext.
    eauto.

    (* empty step *)
    exists Sigma. split*. split*.
    apply* preservation_empty.

    (* pure step *)
    exists Sigma. split*. split*.
    apply* preservation_pure.
  Qed.

  Corollary preservation_many : forall Sigma H P H' P' tr,
    well_typed Sigma H ->
    typing Sigma P ->
    system_step_many H P H' P' tr ->
    exists Sigma',
      extends Sigma Sigma' /\
      well_typed Sigma' H' /\
      typing Sigma' P'.
  Proof.
    introv TypH TypP Step.
    induction* Step.
    specializes* IHStep (>>> __ __).
    destruct IHStep as [ Sigma' [ ? [ ? ? ] ] ].
    forwards Pres : preservation Sigma'; eauto.
    destruct Pres as [ Sigma'' [ ? [ ? ? ] ] ].
    exists* Sigma''.
  Qed.

  Corollary preservation_init : forall P0 H P tr,
    Pres.Typ.typing Pres.Typ.empty_heap_typing P0 ->
    Sys.system_step_many MPres.M.empty_heap P0 H P tr ->
    exists Sigma,
      MPres.well_typed Sigma H /\
      Pres.Typ.typing Sigma P.
  Proof.
    introv Typ Step.
    forwards* Pres : preservation_many.
    apply well_typed_empty.
    destruct* Pres as [ ? [ ? [ ? ? ] ] ].
  Qed.
End System_Model_Type_Preservation.
