Require Import Metatheory.

Require Import maps.
Require Import models.

Set Implicit Arguments.

Open Scope type_scope.

Module Sequential_Consistency (L : LANGUAGE_MODEL) <: MEMORY_MODEL L.
  (* a heap is a map from locations to values *)
  Definition heap := VarMap.t L.val.

  Definition empty_heap : heap := VarMap.empty _.

  Inductive step : heap -> effect L.val -> heap -> Prop :=
  | step_alloc : forall S t r v,
    ~ VarMap.In r S ->
    step S (Some (t, alloc r v)) (VarMap.add r v S)
  | step_read : forall S t r v,
    VarMap.MapsTo r v S ->
    step S (Some (t, read r v)) S
  | step_write : forall S t r v,
    VarMap.In r S ->
    step S (Some (t, write r v)) (VarMap.add r v S)
  | step_spawn : forall H t1 t2,
    step H (Some (t1, spawn _ t2)) H.

  Definition heap_step := step.

  Fact no_pure_steps : forall H H' t,
    ~ (heap_step H (Some (t, pure _)) H').
  Proof.
    intros H H' t Step. inversion Step.
  Qed.

  Inductive heap_step_many : heap -> heap -> trace L.val -> Prop :=
  | heap_step_many_refl : forall H,
    heap_step_many H H nil
  | heap_step_many_trans : forall H1 H2 H3 tr e,
    heap_step_many H1 H2 tr ->
    heap_step H2 e H3 ->
    heap_step_many H1 H3 (e :: tr).
End Sequential_Consistency.
