Require Import Metatheory.

Require Import maps.
Require Import model_facts.
Require Import models.
Require Import msi.
Require Import sc.

Module SCToMSI (L : LANGUAGE_MODEL).
  Module SC := Sequential_Consistency L.
  Module MSI := MSI L.
  Declare Module SysSC : SYSTEM_MODEL L SC.
  Declare Module SysMSI : SYSTEM_MODEL L MSI.
  Module MSIFacts := ModelFacts L MSI SysMSI.

  Inductive map_heap :  SC.heap -> MSI.heap ->Prop :=
  | map_heap_def : forall S C,
    VarPairMap.Empty C ->
    map_heap S (S, C).

  Lemma map_heap_empty : map_heap SC.empty_heap MSI.empty_heap.
  Proof.
    apply map_heap_def.
    intros [ t r ] ?.
    rewrite VarPairMapFacts.find_mapsto_iff.
    rewrite VarPairMapFacts.empty_o.
    intro false.
    inversion false.
  Qed.

  Lemma step_alloc : forall H1 H2 t r v H1',
    map_heap H1 H2 ->
    SC.heap_step H1 (Some (t, alloc r v)) H1' ->
    exists H2',
      MSI.heap_step H2 (Some (t, alloc r v)) H2'
      /\ map_heap H1' H2'.
  Proof with eauto.
    intros H1 H2 t r v H1' MapHeap Step.
    inversion MapHeap; subst.
    inversion Step; subst.
    esplit. split.
    apply MSI.step_alloc...
    apply map_heap_def...
  Qed.

  Lemma map_heap_invalid : forall S C r,
    map_heap S (S, C) ->
    forall t, ~ VarPairMap.In (t, r) C.
  Proof.
    intros S C r MapHeap t [ [ o v ] CacheRead ].
    inversion MapHeap as [ ? ? Emp ]; subst.
    unfold VarPairMap.Empty in Emp.
    specializes Emp (>>> (t, r) (o, v)).
    tauto.
  Qed.

  Lemma step_read : forall H1 H2 t r v H1',
    map_heap H1 H2 ->
    SC.heap_step H1 (Some (t, read r v)) H1' ->
    exists H2', exists tr,
      MSI.heap_step_many H2 H2' tr
      /\ erase_trace tr = (t, read r v) :: nil
      /\ map_heap H1' H2'.
  Proof with eauto.
    intros H1 H2 t r v H1' MapHeap Step.
    inversion MapHeap as [ S C Emp ]; subst.
    inversion Step; subst.
    esplit. esplit. split; [ | split ].
    eapply MSI.heap_step_many_trans.
    eapply MSI.heap_step_many_trans.
    eapply MSI.heap_step_many_trans.
    eapply MSI.heap_step_many_refl.
    apply MSI.step_invalid_to_shared...
    eapply map_heap_invalid...
    apply MSI.step_read_shared.
    rewrite VarPairMapFactsEx2.add_eq_mapsto_iff...
    apply VarPair.eq_refl.
    eapply MSI.step_shared_to_invalid...
    rewrite VarPairMapFactsEx2.add_eq_mapsto_iff...
    apply VarPair.eq_refl.
    simpl...
    apply map_heap_def...
    intros [ t' r' ] [ o' v' ] CacheRead.
    rewrite VarPairMapFacts.remove_mapsto_iff in CacheRead.
    destruct CacheRead as [ Neq CacheRead ].
    rewrite VarPairMapFacts.add_neq_mapsto_iff in CacheRead...
    unfold VarPairMap.Empty in Emp.
    specializes Emp (>>> (t', r') (o', v')).
    tauto.
  Qed.

  Lemma step_write : forall H1 H2 t r v H1',
    map_heap H1 H2 ->
    SC.heap_step H1 (Some (t, write r v)) H1' ->
    exists H2', exists tr,
      MSI.heap_step_many H2 H2' tr
      /\ erase_trace tr = (t, write r v) :: nil
      /\ map_heap H1' H2'.
  Proof with eauto.
    intros H1 H2 t r v H1' MapHeap Step.
    inversion MapHeap as [ S C Emp ]; subst.
    inversion Step as [ | | ? ? ? ? [ ? ? ] | ]; subst.
    esplit. esplit. split; [ | split ].
    eapply MSI.heap_step_many_trans.
    eapply MSI.heap_step_many_trans.
    eapply MSI.heap_step_many_trans.
    eapply MSI.heap_step_many_refl.
    apply MSI.step_invalid_to_modified...
    eapply map_heap_invalid...
    eapply MSI.step_write_modified.
    rewrite VarPairMapFactsEx2.add_eq_mapsto_iff...
    apply VarPair.eq_refl.
    eapply MSI.step_modified_to_invalid...
    rewrite VarPairMapFactsEx2.add_eq_mapsto_iff...
    apply VarPair.eq_refl.
    simpl...
    apply map_heap_def...
    intros [ t' r' ] [ o' v' ] CacheRead.
    rewrite VarPairMapFacts.remove_mapsto_iff in CacheRead.
    destruct CacheRead as [ Neq CacheRead ].
    rewrite VarPairMapFacts.add_neq_mapsto_iff in CacheRead...
    rewrite VarPairMapFacts.add_neq_mapsto_iff in CacheRead...
    unfold VarPairMap.Empty in Emp.
    specializes Emp (>>> (t', r') (o', v')).
    tauto.
  Qed.

  Lemma step_spawn : forall H1 H2 t t' H1',
    map_heap H1 H2 ->
    SC.heap_step H1 (Some (t, spawn _ t')) H1' ->
    exists H2',
      MSI.heap_step H2 (Some (t, spawn _ t')) H2'
      /\ map_heap H1' H2'.
  Proof.
    intros H1 H2 t t' H1' MapHeap Step.
    inversion Step; subst.
    esplit. split.
    apply MSI.step_spawn.
    auto.
  Qed.

  Lemma sc_to_msi_step : forall H1 H2 e H1',
    map_heap H1 H2 ->
    SC.heap_step H1 e H1' ->
    exists H2', exists tr,
      MSI.heap_step_many H2 H2' tr
      /\ erase_trace tr = erase_trace (e :: nil)
      /\ map_heap H1' H2'.
  Proof with eauto.
    intros H1 H2 e H1' MapHeap Step.
    inversion Step; subst.

    forwards [ ? [ ? ? ] ] : step_alloc...
    esplit. esplit. split; [ | split ].
    eapply MSI.heap_step_many_trans...
    auto. auto.

    eapply step_read...

    eapply step_write...

    forwards [ ? [ ? ? ] ] : step_spawn...
    esplit. esplit. split; [ | split ].
    eapply MSI.heap_step_many_trans...
    auto. auto.
  Qed.

  Lemma sc_to_msi_step_many : forall H1 P H2 H1' P' tr,
    map_heap H1 H2 ->
    SysSC.system_step_many H1 P H1' P' tr ->
    exists H2', exists tr',
      SysMSI.system_step_many H2 P H2' P' tr'
      /\ erase_trace tr = erase_trace tr'
      /\ map_heap H1' H2'.
  Proof with eauto.
    intros H1 P H2 H1' P' tr MapHeap StepMany.
    induction StepMany as [ | ? ? ? ? ? ? ? ? ? ? Step ]...

    specializes IHStepMany (>>> __)...
    destruct IHStepMany as [ ? [ ? [ ? [ Erase ? ] ] ] ].
    inversion Step as [ | ? ? ? HeapStep | ]; subst.

    forwards [ ? [ ? [ ? [ Erase' ? ] ] ] ] : sc_to_msi_step...
    esplit. esplit. split; [ | split ].
    eapply MSIFacts.system_step_many_alt...
    eapply MSIFacts.system_step_many_mutual...
    rewrite app_erase.
    rewrite <- Erase.
    rewrite Erase'.
    simpl...
    auto.

    inversion HeapStep.

    esplit. esplit. split; [ | split ].
    eapply SysMSI.system_step_many_trans...
    apply SysMSI.system_step_program...
    simpl.
    rewrite Erase...
    auto.
  Qed.

  Theorem sc_to_msi : forall P0 H1 P tr,
    SysSC.system_step_many SC.empty_heap P0 H1 P tr ->
    exists H2, exists tr',
      SysMSI.system_step_many MSI.empty_heap P0 H2 P tr'.
  Proof with eauto.
    intros P0 H1 P tr StepMany.
    forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : sc_to_msi_step_many...
    apply map_heap_empty.
  Qed.
End SCToMSI.
